from .types import Score


SCORE_INITIALS = {
    'trivial': 'TR',
    'low': 'LO',
    'medium': 'MD',
    'high': 'HI',
    'critical': 'CR',
}
SCORE_NAMES = [
    Score.TRIVIAL.value,
    Score.LOW.value,
    Score.MEDIUM.value,
    Score.HIGH.value,
    Score.CRITICAL.value,
]
